"""Tests the Rotation2d class."""

import math

import pytest
from frccontrol.geometry import Rotation2d


def test_radians_to_degrees():
    rot1 = Rotation2d.from_radians(math.pi / 3)
    rot2 = Rotation2d.from_radians(math.pi / 4)

    assert rot1.degrees == pytest.approx(60.0, abs=1e-9)
    assert rot2.degrees == pytest.approx(45.0, abs=1e-9)


def test_degrees_to_radians():
    rot1 = Rotation2d.from_degrees(45)
    rot2 = Rotation2d.from_degrees(30)

    assert rot1.radians == math.pi / 4
    assert rot2.radians == math.pi / 6


def test_rotate_by_from_zero():
    zero = Rotation2d.from_radians(0)
    rotated = zero + Rotation2d.from_degrees(90)

    assert rotated.radians == math.pi / 2
    assert rotated.degrees == 90.0


def test_rotate_by_non_zero():
    rot = Rotation2d.from_degrees(90)
    rot = rot + Rotation2d.from_degrees(30)

    assert rot.degrees == pytest.approx(120.0, abs=1e-9)


def test_minus():
    rot1 = Rotation2d.from_degrees(70)
    rot2 = Rotation2d.from_degrees(30)

    assert (rot1 - rot2).degrees == pytest.approx(40.0, abs=1e-9)


def test_unary_minus():
    rot = Rotation2d.from_degrees(20)

    assert (-rot).degrees == -20.0


def test_multiply():
    rot = Rotation2d.from_degrees(10)

    assert (rot * 3.0).degrees == pytest.approx(30.0, abs=1e-9)
    assert (rot * 41.0).degrees == pytest.approx(50.0, abs=1e-9)


def test_equality():
    rot1 = Rotation2d.from_degrees(43)
    rot2 = Rotation2d.from_degrees(43)
    assert rot1 == rot2

    rot1 = Rotation2d.from_degrees(-180)
    rot2 = Rotation2d.from_degrees(180)
    assert rot1 == rot2


def test_inequality():
    rot1 = Rotation2d.from_degrees(43)
    rot2 = Rotation2d.from_degrees(43.5)
    assert rot1 != rot2



"""Tests the Twist2d class."""

import math

import pytest
from frccontrol.geometry import Pose2d, Twist2d


def test_straight():
    straight = Twist2d(5, 0, 0)
    straight_transform = straight.exp()

    assert straight_transform.x == 5.0
    assert straight_transform.y == 0.0
    assert straight_transform.rotation.radians == 0.0


def test_quarter_circle():
    quarter_circle = Twist2d(5 / 2 * math.pi, 0, math.pi / 2)
    quarter_circle_transform = quarter_circle.exp()

    assert quarter_circle_transform.x == 5.0
    assert quarter_circle_transform.y == pytest.approx(5.0, abs=1e-9)
    assert quarter_circle_transform.rotation.degrees == 90.0


def test_diagonal_no_dtheta():
    diagonal = Twist2d(2, 2, 0)
    diagonal_transform = diagonal.exp()

    assert diagonal_transform.x == 2.0
    assert diagonal_transform.y == 2.0
    assert diagonal_transform.rotation.degrees == 0.0


def test_equality():
    one = Twist2d(5, 1, 3)
    two = Twist2d(5, 1, 3)
    assert one == two


def test_inequality():
    one = Twist2d(5, 1, 3)
    two = Twist2d(5, 1.2, 3)
    assert one != two


def test_pose2d_log():
    end = Pose2d.from_triplet(5, 5, math.radians(90))
    start = Pose2d.from_triplet(0, 0, 0)

    twist = (end - start).log()

    assert twist == Twist2d(5 / 2 * math.pi, 0, math.pi / 2)

    # Make sure computed twist gives back original end pose
    reapplied = start + twist.exp()
    assert reapplied == end



"""
Steady-state linear Kalman filter class.
"""

import numpy as np
import scipy as sp

from .ctrlutil import covariance_matrix, is_detectable
from .discretization import discretize_ab, discretize_aq, discretize_r


class SteadyStateKalmanFilter:
    """Steady-state linear Kalman filter class."""

    def __init__(
        self,
        plant: sp.signal.StateSpace,
        state_std_devs: list[float],
        measurement_std_devs: list[float],
        dt: float,
    ):
        """
        Constructs a Kalman filter.

        Parameter ``plant``:
            The plant used for the prediction step.
        Parameter ``state_std_devs``:
            Standard deviations of model states.
        Parameter ``measurement_std_devs``:
            Standard deviations of measurements.
        Parameter ``dt``:
            Nominal discretization timestep.
        """
        self.plant = plant

        discA, discQ = discretize_aq(plant.A, covariance_matrix(state_std_devs), dt)
        C = plant.C
        discR = discretize_r(covariance_matrix(measurement_std_devs), dt)

        if not is_detectable(discA, C):
            raise RuntimeError(f"The system is undetectable!\n\nA = {discA}\nC = {C}\n")

        P = sp.linalg.solve_discrete_are(a=discA.T, b=C.T, q=discQ, r=discR)

        # S = CPCᵀ + R
        S = C @ P @ C.T + discR

        # We want to put K = PCᵀS⁻¹ into Ax = b form so we can solve it more
        # efficiently.
        #
        # K = PCᵀS⁻¹
        # KS = PCᵀ
        # (KS)ᵀ = (PCᵀ)ᵀ
        # SᵀKᵀ = CPᵀ
        #
        # The solution of Ax = b can be found via x = A.solve(b).
        #
        # Kᵀ = Sᵀ.solve(CPᵀ)
        # K = (Sᵀ.solve(CPᵀ))ᵀ
        #
        # Drop the transposes on symmetric matrices S and P.
        #
        # K = (S.solve(CP))ᵀ
        self.K = np.linalg.solve(S, C @ P).T

        self.x_hat = np.zeros((self.plant.A.shape[0], 1))

    def reset(self):
        """
        Resets the observer.
        """
        self.x_hat = np.zeros(self.x_hat.shape)

    def predict(self, u, dt):
        """
        Project the model into the future with a new control input u.

        Parameter ``u``:
            New control input from controller.
        Parameter ``dt``:
            Timestep for prediction.
        """
        discA, discB = discretize_ab(self.plant.A, self.plant.B, dt)

        # x̂ₖ₊₁⁻ = Ax̂ₖ⁺ + Buₖ
        self.x_hat = discA @ self.x_hat + discB @ u

    def correct(self, u, y):
        """
        Correct the state estimate x-hat using the measurements in y.

        Parameter ``u``:
            Same control input used in the last predict step.
        Parameter ``y``:
            Measurement vector.
        """
        # x̂ₖ₊₁⁺ = x̂ₖ₊₁⁻ + K(y − (Cx̂ₖ₊₁⁻ + Duₖ₊₁))
        self.x_hat += self.K @ (y - (self.plant.C @ self.x_hat + self.plant.D @ u))



"""Geometry utilities."""

from __future__ import annotations

import math


class Translation2d:
    """Represents a translation in 2D space."""

    def __init__(self, x: float, y: float):
        self.__x = x
        self.__y = y

    @classmethod
    def from_polar(cls, distance: float, angle: Rotation2d) -> Translation2d:
        """
        Constructs a Translation2d with the provided distance and angle. This is
        essentially converting from polar coordinates to Cartesian coordinates.

        Parameter ``distance``:
            The distance from the origin to the end of the translation.
        Parameter ``angle``:
            The angle between the x-axis and the translation vector.
        """
        return Translation2d(distance * angle.cos, distance * angle.sin)

    @property
    def x(self) -> float:
        """Returns x component."""
        return self.__x

    @property
    def y(self) -> float:
        """Returns y component."""
        return self.__y

    def distance(self, other: Translation2d) -> float:
        """
        Calculates the distance between two translations in 2D space.

        The distance between translations is defined as √((x₂−x₁)²+(y₂−y₁)²).

        Parameter ``other``:
            The translation to compute the distance to.
        """
        return math.hypot(other.x - self.x, other.y - self.y)

    def squared_distance(self, other: Translation2d) -> float:
        """
        Calculates the square of the distance between two translations in 2D
        space. This is equivalent to squaring the result of
        distance(Translation2d), but avoids computing a square root.

        The square of the distance between translations is defined as
        (x₂−x₁)²+(y₂−y₁)².

        Parameter ``other``:
            The translation to compute the squared distance to.
        """
        return (other.x - self.x) ** 2 + (other.y - self.y) ** 2

    def norm(self) -> float:
        """Returns the norm, or distance from the origin to the translation."""
        return math.hypot(self.x, self.y)

    def squared_norm(self) -> float:
        """
        Returns the squared norm, or squared distance from the origin to the
        translation. This is equivalent to squaring the result of Norm(), but
        avoids computing a square root.
        """
        return self.x**2 + self.y**2

    def angle(self) -> Rotation2d:
        """Returns the angle this translation forms with the positive X axis."""
        return Rotation2d(self.x, self.y)

    def rotate_by(self, other: Rotation2d) -> Translation2d:
        """
        Applies a rotation to the translation in 2D space.

        Parameter ``other``:
            The rotation by which to rotate.
        """
        return Translation2d(
            self.x * other.cos - self.y * other.sin,
            self.x * other.sin + self.y * other.cos,
        )

    def rotate_around(self, other: Translation2d, rot: Rotation2d) -> Translation2d:
        """
        Rotates this translation around another translation in 2D space.

        Parameter ``other``:
            The other translation to rotate around.
        Parameter ``rot``:
            The rotation to rotate the translation by.
        """
        return Translation2d(
            (self.x - other.x) * rot.cos - (self.y - other.y) * rot.sin + other.x,
            (self.x - other.x) * rot.sin + (self.y - other.y) * rot.cos + other.y,
        )

    def dot(self, other: Translation2d) -> float:
        """
        Computes the dot product between this translation and another
        translation in 2D space.

        The dot product between two translations is defined as x₁x₂+y₁y₂.

        Parameter ``other``:
            The translation to compute the dot product with.
        """
        return self.x * other.x + self.y * other.y

    def cross(self, other: Translation2d) -> float:
        """
        Computes the cross product between this translation and another
        translation in 2D space.

        The 2D cross product between two translations is defined as x₁y₂-x₂y₁.

        Parameter ``other``:
            The translation to compute the cross product with.
        """
        return self.x * other.y - self.y * other.x

    def __add__(self, other: Translation2d) -> Translation2d:
        return Translation2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Translation2d) -> Translation2d:
        return Translation2d(self.x - other.x, self.y - other.y)

    def __neg__(self) -> Translation2d:
        return Translation2d(-self.x, -self.y)

    def __mul__(self, scalar: float) -> Translation2d:
        return Translation2d(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar: float) -> Translation2d:
        return Translation2d(self.x / scalar, self.y / scalar)

    def __eq__(self, other) -> bool:
        return abs(self.x - other.x) < 1e-9 and abs(self.y - other.y) < 1e-9


class Rotation2d:
    """Represents a rotation in 2D space."""

    def __init__(self, x: float, y: float):
        magnitude = math.hypot(x, y)
        if magnitude > 1e-6:
            self.__cos = x / magnitude
            self.__sin = y / magnitude
        else:
            raise RuntimeError("x and y components of Rotation2d are zero")

    @classmethod
    def from_radians(cls, heading: float) -> Rotation2d:
        """
        Constructs a Rotation2d from a heading in radians.

        Parameter ``heading``:
            The heading in radians.
        """
        return cls(math.cos(heading), math.sin(heading))

    @classmethod
    def from_degrees(cls, heading: float) -> Rotation2d:
        """
        Constructs a Rotation2d from a heading in degrees.

        Parameter ``heading``:
            The heading in degrees.
        """
        return cls(math.cos(math.radians(heading)), math.sin(math.radians(heading)))

    @property
    def radians(self) -> float:
        """Returns the heading in radians."""
        return math.atan2(self.sin, self.cos)

    @property
    def degrees(self) -> float:
        """Returns the heading in degrees."""
        return math.degrees(math.atan2(self.sin, self.cos))

    @property
    def cos(self) -> float:
        """Returns the cosine of the heading."""
        return self.__cos

    @property
    def sin(self) -> float:
        """Returns the sine of the heading."""
        return self.__sin

    @property
    def tan(self) -> float:
        """Returns the tangent of the heading."""
        return self.sin / self.cos

    def __add__(self, other: Rotation2d) -> Rotation2d:
        return self.rotate_by(other)

    def __sub__(self, other: Rotation2d) -> Rotation2d:
        return self.rotate_by(-other)

    def __neg__(self) -> Rotation2d:
        return Rotation2d(self.cos, -self.sin)

    def rotate_by(self, other: Rotation2d) -> Rotation2d:
        """
        Rotates the rotation in 2D space.

        Parameter ``other``:
            The rotation by which to rotate.
        """
        return Rotation2d(
            self.cos * other.cos - self.sin * other.sin,
            self.cos * other.sin + self.sin * other.cos,
        )

    def __mul__(self, scalar: float) -> Rotation2d:
        return Rotation2d.from_radians(self.radians * scalar)

    def __truediv__(self, scalar: float) -> Rotation2d:
        return Rotation2d.from_radians(self.radians / scalar)

    def __eq__(self, other) -> bool:
        return math.hypot(self.cos - other.cos, self.sin - other.sin) < 1e-9


class Pose2d:
    """Represents a pose in 2D space."""

    def __init__(self, translation: Translation2d, rotation: Rotation2d):
        self.__translation = translation
        self.__rotation = rotation

    @classmethod
    def from_triplet(cls, x: float, y: float, θ: float) -> Pose2d:
        """
        Constructs a Pose2d from an x-y-heading triplet.

        Parameter ``x``:
            The x component.
        Parameter ``y``:
            The y component.
        Parameter ``θ``:
            The heading component.
        """
        return cls(Translation2d(x, y), Rotation2d.from_radians(θ))

    @property
    def translation(self) -> Translation2d:
        """Returns the translation."""
        return self.__translation

    @property
    def rotation(self) -> Rotation2d:
        """Returns the rotation."""
        return self.__rotation

    @property
    def x(self) -> float:
        """Returns the x component."""
        return self.translation.x

    @property
    def y(self) -> float:
        """Returns the y component."""
        return self.translation.y

    def __add__(self, other: Transform2d) -> Pose2d:
        return self.transform_by(other)

    def __sub__(self, other: Pose2d) -> Transform2d:
        pose = self.relative_to(other)
        return Transform2d(pose.translation, pose.rotation)

    def rotate_by(self, other: Rotation2d) -> Pose2d:
        """
        Rotates the pose in 2D space.

        Parameter ``other``:
            The rotation by which to rotate.
        """
        return Pose2d(self.translation.rotate_by(other), self.rotation.rotate_by(other))

    def transform_by(self, other: Transform2d) -> Pose2d:
        """
        Transforms this pose by the given transformation.

        Parameter ``other``:
            The transformation by which to transform.
        """
        return Pose2d(
            self.translation + other.translation.rotate_by(self.rotation),
            other.rotation + self.rotation,
        )

    def relative_to(self, other: Pose2d) -> Pose2d:
        """
        Returns this pose relative to another pose.

        Parameter ``other``:
            The origin for the relative pose.
        """
        transform = Transform2d.from_poses(other, self)
        return Pose2d(transform.translation, transform.rotation)

    def rotate_around(self, point: Translation2d, rot: Rotation2d) -> Pose2d:
        """
        Rotates the pose around a point in 2D space.

        Parameter ``point``:
            The point in 2D space to rotate around.
        Parameter ``rot``:
            The rotation to rotate the pose by.
        """
        return Pose2d(
            self.translation.rotate_around(point, rot), self.rotation.rotate_by(rot)
        )

    def __eq__(self, other) -> bool:
        return (
            self.__translation == other.__translation
            and self.__rotation == other.__rotation
        )


class Transform2d:
    """Represents a transformation of a pose in the pose's frame."""

    def __init__(self, translation: Translation2d, rotation: Rotation2d):
        self.__translation = translation
        self.__rotation = rotation

    @classmethod
    def from_triplet(cls, x: float, y: float, θ: float) -> Transform2d:
        """
        Constructs a Transform2d from an x-y-heading triplet.

        Parameter ``x``:
            The x component.
        Parameter ``y``:
            The y component.
        Parameter ``θ``:
            The heading component.
        """
        return cls(Translation2d(x, y), Rotation2d.from_radians(θ))

    @classmethod
    def from_poses(cls, initial: Pose2d, final: Pose2d) -> Transform2d:
        """
        Constructs a Transform2d from initial and final poses.

        Parameter ``initial``:
            The initial pose.
        Parameter ``final``:
            The final pose.
        """
        return cls(
            (final.translation - initial.translation).rotate_by(-initial.rotation),
            final.rotation - initial.rotation,
        )

    @property
    def translation(self) -> Translation2d:
        """Returns the translation."""
        return self.__translation

    @property
    def rotation(self) -> Rotation2d:
        """Returns the rotation."""
        return self.__rotation

    @property
    def x(self) -> float:
        """Returns the x component."""
        return self.translation.x

    @property
    def y(self) -> float:
        """Returns the y component."""
        return self.translation.y

    def log(self) -> Twist2d:
        """Takes the log of the transformation to obtain a twist."""
        dtheta = self.rotation.radians
        half_dtheta = dtheta / 2

        cos_minus_one = self.rotation.cos - 1

        if abs(cos_minus_one) < 1e-9:
            half_theta_by_tan_of_half_dtheta = 1.0 - 1.0 / 12.0 * dtheta * dtheta
        else:
            half_theta_by_tan_of_half_dtheta = (
                -(half_dtheta * self.rotation.sin) / cos_minus_one
            )

        translation_part = self.translation.rotate_by(
            Rotation2d(half_theta_by_tan_of_half_dtheta, -half_dtheta)
        ) * math.hypot(half_theta_by_tan_of_half_dtheta, half_dtheta)

        return Twist2d(translation_part.x, translation_part.y, dtheta)

    def inverse(self) -> Transform2d:
        """
        Invert the transformation. This is useful for undoing a transformation.
        """
        return Transform2d(
            (-self.translation).rotate_by(-self.rotation), -self.rotation
        )

    def __mul__(self, scalar: float) -> Transform2d:
        return Transform2d(self.translation * scalar, self.rotation * scalar)

    def __truediv__(self, scalar: float) -> Transform2d:
        return self * (1.0 / scalar)

    def __add__(self, other: Transform2d) -> Transform2d:
        """
        Composes two transformations. The second transform is applied relative
        to the orientation of the first.

        Parameter ``other``:
            The transform to compose with this one.
        """
        return Transform2d.from_poses(
            Pose2d.from_triplet(0, 0, 0),
            Pose2d.from_triplet(0, 0, 0).transform_by(self).transform_by(other),
        )

    def __eq__(self, other) -> bool:
        return (
            self.__translation == other.__translation
            and self.__rotation == other.__rotation
        )


class Twist2d:
    """A change in distance in the tangent space of 2D pose."""

    def __init__(self, dx: float, dy: float, dθ: float):
        """
        Constructs a Twist2d.

        Parameter ``dx``:
            Linear "dx" component.
        Parameter ``dy``:
            Linear "dy" component.
        Parameter ``dθ``:
            Angular "dtheta" component (radians).
        """
        self.__dx = dx
        self.__dy = dy
        self.__dθ = dθ

    @property
    def dx(self) -> float:
        """Returns the dx component."""
        return self.__dx

    @property
    def dy(self) -> float:
        """Returns the dy component."""
        return self.__dy

    @property
    def dθ(self) -> float:
        """Returns the dθ component."""
        return self.__dθ

    def exp(self) -> Transform2d:
        """Exponentiates the twist to obtain a transformation."""
        rot = Rotation2d.from_radians(self.dθ)

        if abs(self.dθ) < 1e-9:
            s = 1 - 1 / 6 * self.dθ * self.dθ
            c = 0.5 * self.dθ
        else:
            s = rot.sin / self.dθ
            c = (1 - rot.cos) / self.dθ

        return Transform2d(
            Translation2d(self.dx * s - self.dy * c, self.dx * c + self.dy * s), rot
        )

    def __eq__(self, other) -> bool:
        return (
            abs(self.dx - other.dx) < 1e-9
            and abs(self.dy - other.dy) < 1e-9
            and abs(self.dθ - other.dθ) < 1e-9
        )



"""Tests the Translation2d class."""

import math

import pytest
from frccontrol.geometry import Rotation2d, Translation2d


def test_sum():
    one = Translation2d(1, 3)
    two = Translation2d(2, 5)

    sum = one + two

    assert sum.x == 3.0
    assert sum.y == 8.0


def test_difference():
    one = Translation2d(1, 3)
    two = Translation2d(2, 5)

    difference = one - two

    assert difference.x == -1.0
    assert difference.y == -2.0


def test_rotate_by():
    another = Translation2d(3, 0)
    rotated = another.rotate_by(Rotation2d.from_degrees(90))

    assert rotated.x == pytest.approx(0.0, abs=1e-9)
    assert rotated.y == pytest.approx(3.0, abs=1e-9)


def test_rotate_around():
    translation = Translation2d(2, 1)
    other = Translation2d(3, 2)
    rotated = translation.rotate_around(other, Rotation2d.from_degrees(180))

    assert rotated.x == pytest.approx(4.0, abs=1e-9)
    assert rotated.y == pytest.approx(3.0, abs=1e-9)


def test_multiplication():
    original = Translation2d(3, 5)
    mult = original * 3

    assert mult.x == 9.0
    assert mult.y == 15.0


def test_division():
    original = Translation2d(3, 5)
    div = original / 2

    assert div.x == 1.5
    assert div.y == 2.5


def test_norm():
    one = Translation2d(3, 5)
    assert one.norm() == math.hypot(3.0, 5.0)


def test_squared_norm():
    one = Translation2d(3, 5)
    assert one.squared_norm() == 34.0


def test_distance():
    one = Translation2d(1, 1)
    two = Translation2d(6, 6)
    assert one.distance(two) == 5.0 * math.sqrt(2.0)


def test_squared_distance():
    one = Translation2d(1, 1)
    two = Translation2d(6, 6)
    assert one.squared_distance(two) == 50.0


def test_unary_minus():
    original = Translation2d(-4.5, 7)
    inverted = -original

    assert inverted.x == 4.5
    assert inverted.y == -7.0


def test_equality():
    one = Translation2d(9, 5.5)
    two = Translation2d(9, 5.5)
    assert one == two


def test_inequality():
    one = Translation2d(9, 5.5)
    two = Translation2d(9, 5.7)
    assert one != two


def test_polar_constructor():
    one = Translation2d.from_polar(math.sqrt(2), Rotation2d.from_degrees(45))
    assert one.x == pytest.approx(1.0, abs=1e-9)
    assert one.y == 1.0

    two = Translation2d.from_polar(2, Rotation2d.from_degrees(60))
    assert two.x == pytest.approx(1.0, abs=1e-9)
    assert two.y == math.sqrt(3.0)


def test_dot():
    one = Translation2d(2, 3)
    two = Translation2d(3, 4)
    assert one.dot(two) == 18.0


def test_cross():
    one = Translation2d(2, 3)
    two = Translation2d(3, 4)
    assert one.cross(two) == -1.0



"""Tests the Pose2d class."""

import math

import pytest
from frccontrol.geometry import Pose2d, Rotation2d, Transform2d, Translation2d


def test_rotate_by():
    x = 1
    y = 2
    initial = Pose2d.from_triplet(x, y, math.radians(45))

    rotation = Rotation2d.from_degrees(5)
    rotated = initial.rotate_by(rotation)

    # Translation is rotated by CCW rotation matrix
    c = rotation.cos
    s = rotation.sin
    assert rotated.x == c * x - s * y
    assert rotated.y == s * x + c * y
    assert rotated.rotation.degrees == pytest.approx(
        initial.rotation.degrees + rotation.degrees, abs=1e-9
    )


def test_transform_by():
    initial = Pose2d.from_triplet(1, 2, math.radians(45))
    transform = Transform2d(Translation2d(5, 0), Rotation2d.from_degrees(5))

    transformed = initial + transform

    assert transformed.x == 1 + 5 / math.sqrt(2)
    assert transformed.y == 2 + 5 / math.sqrt(2)
    assert transformed.rotation.degrees == pytest.approx(50.0, abs=1e-9)


def test_relative_to():
    initial = Pose2d.from_triplet(0, 0, math.radians(45))
    final = Pose2d.from_triplet(5, 5, math.radians(45))

    final_relative_to_initial = final.relative_to(initial)

    assert final_relative_to_initial.x == pytest.approx(5.0 * math.sqrt(2.0), abs=1e-9)
    assert final_relative_to_initial.y == pytest.approx(0.0, abs=1e-9)
    assert final_relative_to_initial.rotation.degrees == pytest.approx(0.0, abs=1e-9)


def test_rotate_around():
    initial = Pose2d.from_triplet(5, 0, 0)
    point = Translation2d(0, 0)

    rotated = initial.rotate_around(point, Rotation2d.from_degrees(180))

    assert rotated.x == pytest.approx(-5.0, abs=1e-9)
    assert rotated.y == pytest.approx(0.0, abs=1e-9)
    assert rotated.rotation.degrees == pytest.approx(180.0, abs=1e-9)


def test_equality():
    a = Pose2d.from_triplet(0, 5, math.radians(43))
    b = Pose2d.from_triplet(0, 5, math.radians(43))
    assert a == b


def test_inequality():
    a = Pose2d.from_triplet(0, 5, math.radians(43))
    b = Pose2d.from_triplet(0, 1.524, math.radians(43))
    assert a != b


def test_minus():
    initial = Pose2d.from_triplet(0, 0, math.radians(45))
    final = Pose2d.from_triplet(5, 5, math.radians(45))

    transform = final - initial

    assert transform.x == pytest.approx(5.0 * math.sqrt(2.0), abs=1e-9)
    assert transform.y == pytest.approx(0.0, abs=1e-9)
    assert transform.rotation.degrees == pytest.approx(0.0, abs=1e-9)



"""Tests the Transform2d class."""

import math

from frccontrol.geometry import Pose2d, Transform2d


def test_inverse():
    initial = Pose2d.from_triplet(1, 2, math.radians(45))
    transform = Transform2d.from_triplet(5, 0, math.radians(5))

    transformed = initial + transform
    untransformed = transformed + transform.inverse()

    assert untransformed == initial


def test_composition():
    initial = Pose2d.from_triplet(1, 2, math.radians(45))
    transform1 = Transform2d.from_triplet(5, 0, math.radians(5))
    transform2 = Transform2d.from_triplet(0, 2, math.radians(5))

    transformed_separate = initial + transform1 + transform2
    transformed_combined = initial + (transform1 + transform2)

    assert transformed_separate == transformed_combined
